unit Hooks;

interface

uses
  Windows, Messages, SysUtils, Classes, RSSysUtils, RSQ, Common, RSCodeHook,
  Math, MP3, RSDebug, IniFiles, Direct3D, MMSystem, Graphics;

procedure HookAll;

implementation

var
  QuickSaveUndone, Autorun: Boolean;
  DoubleSpeed: BOOL;

procedure ProcessMouseLook; forward;

//----- Functions

procedure AddAction(action, info1, info2:int); stdcall;
type
  PActionQueueItem = ^TActionQueueItem;
  TActionQueueItem = packed record
    Action: int;
    Info1: int;
    Info2: int;
  end;
  PActionQueue = ^TActionQueue;
  TActionQueue = packed record
    Count: int;
    Items: array[0..39] of TActionQueueItem;
  end;

const
  Queue = PActionQueue(_ActionQueue);
begin
  with Queue^ do
    if Count < 40 then
    begin
      Items[Count]:= PActionQueueItem(@action)^;
      inc(Count);
    end;
end;

function GetCurrentMember:ptr;
begin
  Result:= pptr(_PartyMembers + 4*_CurrentMember^)^;
end;

procedure DoPlaySound;
asm
  mov ecx, _PlaySoundStruct
  push _PlaySound
end;

var
  PlaySound: procedure(id: int; a3: int = 0; a4: int = 0; a5: int = -1; a6: int = 0; a7: int = 0; a8: int = 0; a9: int = 0); stdcall;

procedure ShowStatusText(text: string; time: int = 2);
begin
  _ShowStatusText(0, time, ptr(text));
  _NeedRedraw^:= 1;
end;

//----- Fix keys configuration loading

const
  DefaultKeys: array[0..29] of int = (38, 40, 37, 39, 89, 88, 13, 83, 65, 32, 67, 66, 9, 81, 90, 82, 84, 78, 77, 85, 34, 46, 35, 107, 109, 33, 45, 36, 219, 221);
  _KeysPtr = ptr($69AC8C);
var
  KeyConfigDone: Boolean;
  StoredKeys: array[0..29] of int;

procedure FixKeyConfigProc;
begin
  CopyMemory(@StoredKeys, _KeysPtr, SizeOf(StoredKeys));
  CopyMemory(_KeysPtr, @DefaultKeys, SizeOf(DefaultKeys));
end;

procedure FixKeyConfigHook;
asm
  call FixKeyConfigProc
  pop edi
  pop esi
  pop ebx
  leave
end;

procedure DoKeyConfig;
const
  UpdateShortCut: procedure(a1, new, old: int; first: BOOL) = ptr($41B48A);
var
  first: BOOL;
  i: int;
begin
  CopyMemory(_KeysPtr, @StoredKeys, SizeOf(StoredKeys));
  first:= true;
  for i := 0 to high(DefaultKeys) do
    if (StoredKeys[i] <> DefaultKeys[i]) then
    begin
      UpdateShortCut(0, StoredKeys[i], DefaultKeys[i], first);
      first:= false;
    end;
end;

//----- Keys

var
  KeysChecked: array[0..255] of Boolean;

function MyGetAsyncKeyState(vKey: Integer): SHORT; stdcall;
const
  PressedBit: array[Boolean] of SHORT = (1, 0);
begin
  vKey:= vKey and $ff;
  Result:= GetAsyncKeyState(vKey);
  if (Result < 0) and not KeysChecked[vKey] then
    Result:= Result or 1;
  KeysChecked[vKey]:= Result < 0;
end;

function CheckKey(key: int):Boolean;
begin
  Result:= (MyGetAsyncKeyState(key) and 1) <> 0;
end;

//----- Quick Save, keys related things

var
  LastMusicStatus, MusicLoopsLeft: int;

procedure KeysProc;
var
  nopause: Boolean;
  status, loops: int;
begin
  // Don't allow using inactive members
  if (_Paused^=0) and (_CurrentScreen^ = 0) and (_CurrentMember^ <> 0) and
     (pword(int(GetCurrentMember) + _CharOff_Recover)^ > 0) and
     FixInactivePlayersActing then
  begin
    _CurrentMember^:= _FindActiveMember;
    _NeedRedraw^:= 1;
  end;

  // Fix keys configuration loading
  if not KeyConfigDone then
    DoKeyConfig;
  KeyConfigDone:= true;

  // MusicLoopsCount
  if MusicLoopsLeft <> 0 then
  begin
    status:= RedbookStatus;
    if (status = 3) and (LastMusicStatus in [1, 2]) then
    begin
      loops:= MusicLoopsLeft;
      _LoadMapTrack;
      if loops > 0 then
        MusicLoopsLeft:= loops - 1;
    end;
    LastMusicStatus:= status;
  end;

  nopause:= (_Paused^=0) and (_CurrentScreen^ = 0);

  // QuickSave
  if (QuickSavesCount >= 0) and CheckKey(QuickSaveKey) and nopause then
  begin
    QuickSaveUndone:= true;
    _DoSaveGame(0, 0, 2);
    if QuickSaveUndone then
      PlaySound(27);
  end;

  // QuickLoad
  if (QuickSavesCount >= 0) and CheckKey(Options.QuickLoadKey) and nopause
     and FileExists('saves\quiksave.mm7') then
  begin
    _Paused^:= 1;
    pint($69CDA8)^:= 1;
    _SaveSlotsFiles[0]:= 'quiksave.mm7';
    _DoLoadGame(0, 0, 0);
    pint($6A0BC8)^:= 3;
  end;

  // InventoryKey
  if CheckKey(Options.InventoryKey) and nopause then
  begin
    if _CurrentMember^ = 0 then
      _CurrentMember^:= 1;
    _CurrentCharScreen^:= 103;
    _OpenInventory_result^:= _OpenInventory_part;
    _NeedRedraw^:= 1;
  end;

  // CharScreenKey
  if CheckKey(Options.CharScreenKey) then
    case _CurrentScreen^ of
      0:
        if _Paused^=0 then
        begin
          if _CurrentMember^ = 0 then
            _CurrentMember^:= 1;
          _OpenInventory_result^:= _OpenInventory_part;
          _NeedRedraw^:= 1;
        end;
      7:
        AddAction(113, BoolToInt[_EscKeyUnkCheck^ <> 0], 0);
    end;

  // DoubleSpeedKey
  if CheckKey(Options.DoubleSpeedKey) then
  begin
    DoubleSpeed:= not DoubleSpeed;
    if DoubleSpeed then
      ShowStatusText(SDoubleSpeed)
    else
      ShowStatusText(SNormalSpeed);
  end;

  // Autorun like in WoW
  if CheckKey(Options.AutorunKey) then
    Autorun:= not Autorun;

  // MouseLookChangeKey
  if _CurrentScreen^ <> 0 then
    MouseLookChanged:= false
  else if CheckKey(Options.MouseLookChangeKey) then
    MouseLookChanged:= not MouseLookChanged;

  // MouseLook
  if Options.MouseLook then
    ProcessMouseLook;
end;

procedure KeysHook;
asm
  call KeysProc
  jmp GetAsyncKeyState
end;

//----- Buggy autosave file name localization

var
  SaveNamesStd: procedure;

procedure SaveNamesHook;
begin
  SaveNamesStd;
  _AutosaveFile^:= 'autosave.mm7';
end;

//----- Fix Save/Load Slots

var
  SaveName: string;
  SaveSpecial: string = 'quiksave.mm7';
  FillSaveSlotsStd: ptr;

procedure FillSaveSlotsBefore;
var
  name: PChar;
begin
  name:= @_SaveSlotsFiles^[_SaveSlot^];
  if (name^ = #0) or (_SaveSlotsCount^ = 0) then  exit;

  if CompareMem(name, PChar('save'), 4) then
    SaveName:= name
  else
    if SaveSpecial = '' then
      SaveSpecial:= name;
end;

procedure FillSaveSlotsAfter(save: bool);
var
  slot, count: int;
  name: string;
begin
  name:= SaveName;
  if not save and (SaveSpecial <> '') then
    name:= SaveSpecial;

  if save then
    count:= 40
  else
    count:= _SaveSlotsCount^;

  if name <> '' then
  begin
    slot:= max(count - 1, 0);
    while (slot > 0) and (AnsiStrComp(_SaveSlotsFiles^[slot], ptr(name)) <> 0) do
      dec(slot);
  end else
    slot:= 0;

  _SaveSlot^:= slot;
  if (slot < _SaveScroll^) or (slot >= _SaveScroll^ + 7) then
    _SaveScroll^:= slot - 3;
  _SaveScroll^:= max(0, min(_SaveScroll^, count - 7));
  if not save then
    SaveSpecial:= '';
end;

procedure FillSaveSlotsHook;
asm
  push ecx
  call FillSaveSlotsBefore
  mov ecx, [esp]
  call FillSaveSlotsStd
  pop eax
  call FillSaveSlotsAfter
end;

procedure ChooseSaveSlotProc;
begin
  SaveSpecial:= '';
end;

var
  ChooseSaveSlotBackup: ptr;

procedure ChooseSaveSlotHook;
asm
  call ChooseSaveSlotProc
  jmp ChooseSaveSlotBackup
end;

var
  SaveBackup: ptr;

procedure SaveHook;
asm
  call ChooseSaveSlotProc
  jmp SaveBackup
end;

//----- Quicksave trigger

procedure QuicksaveProc;
var
  i: int;
  s1, s2: string;
begin
  ShowStatusText(_TextGameSaved^);
  QuickSaveUndone:= false;
  SaveSpecial:= 'quiksave.mm7';
  if QuickSavesCount <= 1 then
    exit;
  s1:= Format('saves\quiksave%d.mm7', [QuickSavesCount]);
  DeleteFile(s1);
  for i := QuickSavesCount - 1 downto 2 do
  begin
    s2:= s1;
    s1:= Format('saves\quiksave%d.mm7', [i]);
    MoveFile(ptr(s1), ptr(s2));
    DeleteFile(s1);
  end;
  MoveFile('saves\quiksave.mm7', ptr(s1));
end;

procedure QuicksaveHook;
const
  QuiksaveStr: PChar = 'saves\quiksave.mm7';
asm
  cmp [ebp - $24], 2
  jnz @norm

  push eax
  call QuicksaveProc
  pop eax

  pop ecx
  push QuiksaveStr
  jmp ecx
  
@norm:
  pop ecx
  push $4E95B4
  jmp ecx
end;

//----- Show multiple Quick Saves

procedure QuickSaveSlotProc;
var
  i: int;
  name, s: string;
begin
  s:= '';
  for i := 1 to QuickSavesCount do
  begin
    if i <> 1 then
      s:= IntToStr(i);
    name:= Format('quiksave%s.mm7', [s]);
    MoveFile(ptr(Format('quick%s.mm7', [s])), ptr(name));
    if _access(ptr(name)) <> -1 then
    begin
      StrCopy(_SaveSlotsFiles^[_SaveSlotsCount^], ptr(name));
      inc(_SaveSlotsCount^);
    end;
  end;
end;

procedure QuickSaveSlotHook;
asm
  mov esi, $5C5C30
  cmp [esp + $14], 0
  jnz @exit
  call QuickSaveSlotProc
@exit:
end;

var
  QuickSaveNamesTmp: string;

function QuickSaveNamesProc(fileName: PChar): bool;
var
  name: string;
begin
  Result:= CompareMem(fileName, PChar('quiksave'), 8);
  if Result then
  begin
    name:= fileName;
    name:= Copy(name, 9, length(name) - 12);
    if name <> '' then
      name:= QuickSaveDigitSpace + name;
    QuickSaveNamesTmp:= #2 + name;
  end else
    if CompareMem(fileName, PChar('autosave'), 8) then
    begin
      Result:= true;
      QuickSaveNamesTmp:= #1;
    end;
end;

procedure QuickSaveNamesHook;
const
  AutosaveFile = int(_AutosaveFile);
asm
  mov eax, [esp + $1C]
  call QuickSaveNamesProc
  test eax, eax
  pop eax
  jnz @quik

  push dword ptr [AutosaveFile]
  jmp eax

@quik:
  push QuickSaveNamesTmp
  push $45E676
end;

function QuickSaveDrawProc(name: PChar): PChar;
begin
  case name^ of
    #1:
      Result:= _AutosaveName^;
    #2:
    begin
      QuickSaveNamesTmp:= QuickSaveName + (name + 1);
      Result:= ptr(QuickSaveNamesTmp);
    end;
    else
      Result:= name;
  end;
end;

procedure QuickSaveDrawHook;
asm
  push ecx
  push eax
  mov eax, edi
  call QuickSaveDrawProc
  mov [esp + 12], eax
  pop eax
  pop ecx
  mov edx, [$5C3488]
end;

procedure QuickSaveDrawHook2;
asm
  push ecx
  lea eax, $69BBE8[eax]
  call QuickSaveDrawProc
  mov edx, eax
  pop ecx
end;

//----- Autosave didn't pause time

var
  SaveSetPauseStd: function(a1, a2, a3, b4, b5, b6: int): int;

function SaveSetPauseHook(a1, a2, a3, b4, b5, b6: int): int;
var
  old: int;
begin
  old:= _Paused^;
  _Paused^:= 1;
  Result:= SaveSetPauseStd(a1, a2, a3, b4, b5, b6);
  _Paused^:= old;
end;

//----- CapsLockToggleRun

procedure CapsLockHook;
asm
  xor eax, eax
  ret 4
end;

//----- NoIntro

procedure IntroHook;
asm
  mov ecx, $4EFE74 // "Intro"
  push $4BE671 // ShowMovie
end;

//----- Multiple fauntain drinks bug

function InactiveMemberEventsProc: int;
type
  TMembersArray = array[0..4] of ptr;
  PMembersArray = ^TMembersArray;
const
  Members = PMembersArray(_PartyMembers);
var
  a: array[0..3] of byte;
  i, n: int;
begin
  n:= 0;
  for i := 1 to 4 do
    if _Character_IsAlive(nil, nil, Members^[i]) then
    begin
      a[n]:= i;
      inc(n);
    end;
  if n > 0 then
    Result:= a[Random(n)]
  else
    Result:= 1;
end;

procedure InactiveMemberEventsHook;
const
  CurMember = int(_CurrentMember);
asm
  mov eax, [CurMember]
  test eax, eax
  pop eax
  jnz @normal
// temporary set CurrentMember to random value, but fixed during the whole event
  pushad
  call InactiveMemberEventsProc
  mov [CurMember], eax
  popad
  push [esp + 4]
  push offset @after
@normal:
  sub esp, 498h
  inc eax
  jmp eax
@after:
  xor eax, eax
  mov [CurMember], eax
  ret 4
end;

//----- Switch between inactive members in inventory screen

procedure CheckNextMemberHook;
const
  CurScreen = int(_CurrentScreen);
asm
  cmp [CurScreen], 7
  jz @ok
  cmp word ptr [esi + _CharOff_Recover], 0
@ok:
end;

//----- Show recovery time in Attack/Shoot description

function AttackDescriptionProc(str: PChar; shoot: LongBool):PChar;
var
  member: PChar;
  i: int;
  blaster: Boolean;
begin
  blaster:= false;
  member:= GetCurrentMember;
  i:= pint(member + _CharOff_ItemMainHand)^;
  if (i >= 0) then
  begin
    i:= int(member + _CharOff_Items + i*_ItemOff_Size);
    blaster:= (pint(i + $14)^ and 2 = 0) and (pbyte(_ItemsTxt + $30*pint(i)^ + $1D)^ = 7);
  end;
  i:= _Character_GetWeaponDelay(0, 0, member, shoot and not blaster);
  if (i < 30) and not (shoot or blaster) then
    i:= 30;
  StrLCopy(_TextBuffer2, str, 499);
  Result:= StrLCat(_TextBuffer2, ptr(Format(RecoveryTimeInfo, [i])), 499);
end;

procedure AttackDescriptionHook;
asm
  xor edx, edx
  mov eax, ebx
  call AttackDescriptionProc
  mov ebx, eax
  push $418437
end;

procedure ShootDescriptionHook;
asm
  xor edx, edx
  inc edx
  mov eax, ebx
  call AttackDescriptionProc
  mov ebx, eax
  push $418437
end;

//----- Load files from DataFiles folder

function LodFilesProc(name: PChar; UseMalloc: Boolean): ptr;
var
  f: HFILE;
  n: DWORD;
begin
  Result:= nil;
  f:= FileOpen('DataFiles\' + name, fmOpenRead);
  if f = INVALID_HANDLE_VALUE then  exit;

  try
    n:= GetFileSize(f, nil);
    if UseMalloc then
      Result:= _malloc(n)
    else
      Result:= _Alloc(0, 0, _Allocator, name, n, 0);
    FileRead(f, Result^, n);
  finally
    FileClose(f);
  end
end;

procedure LodFilesHook;
const
  lod = $6BE8D8; // events.lod
asm
  cmp ecx, lod
  jnz @std
  mov eax, [esp + 8]
  mov edx, [esp + 12]
  push ecx
  call LodFilesProc
  pop ecx
  test eax, eax
  jz @std
  pop ecx
  ret 8

@std:
  pop eax
  push ebp
  mov ebp, esp
  sub esp, $4C
  jmp eax
end;

//----- Dagger tripple damage from 2nd hand check

procedure SecondHandDaggerHook;
asm
  xor eax, eax
  cmp word ptr [edi + $10C], 128
  jb @exit
  inc eax
@exit:
end;

//----- Include items bonus to skill in recovery time

procedure WeaponRecoveryBonusHook;
asm
  cmp ecx, 64
end;

//----- Don't allow using scrolls by not active member

var
  ScrollTestMemberStd: ptr;

function FindActiveMember: int;
begin
  Result:= _FindActiveMember;
end;

procedure ScrollTestMember;
const
  TurnBased = int(_TurnBased);
asm
  push ecx
  mov eax, ScrollTestMemberStd
  call eax
  pop ecx
  test eax, eax
  jz @exit
  cmp word ptr [ecx + _CharOff_Recover], 0 // recovering
  jnz @beep
  cmp dword ptr [TurnBased], 0
  jz @exit
  push ecx
  call FindActiveMember
  pop ecx
  imul eax, $1B3C
  add eax, $ACBCC8
  cmp ecx, eax
  jz @exit
@beep:
  pop eax
  mov ecx, PlayerNotActive
  mov edx, 2
  push $46861D
@exit:
end;

//----- Show numerical reputation value

var
  ReputationHookStd: ptr;

procedure ReputationHook;
asm
  push ecx
  call ReputationHookStd
  pop ecx
  xchg eax, [esp]
  push eax
  mov eax, ecx
end;

//----- DoubleSpeed

var
  DoubleSpeedStd: ptr;

procedure DoubleSpeedHook;
asm
  cmp DoubleSpeed, 0
  jz @norm
  mov eax, [esi + $1C]
  shl eax, 1
  mov [esi + $1C], eax
@norm:
  call DoubleSpeedStd
end;

//----- TurnSpeed

var
  TurnSpeedStd: ptr;

procedure TurnSpeedHook;
asm
  cmp DoubleSpeed, 0
  jz @norm
  fmul TurnSpeedDouble
  jmp TurnSpeedStd
@norm:
  fmul TurnSpeedNormal
  jmp TurnSpeedStd
end;

//----- fix 'of Feather Falling' items

var
  FeatherFallStd: ptr;

procedure FeatherFallHook;
asm
  push 72
  call _Character_WearsItemWithEnchantSpec
  test eax, eax
  jnz @feather

  push 8
  push 529
  mov ecx, [esi]
  call _Character_WearsItem
  test eax, eax
  jnz @feather

  mov ecx, [esi]
  jmp FeatherFallStd

@feather:
  push $474FF3
  ret 8
end;

//----- ProgressiveDaggerTrippleDamage

procedure DaggerTrippleHook;
asm
  idiv ecx
  push edx
  mov ecx, edi
  push 2
  call _Character_GetSkillWithBonuses
  and eax, $3F
  pop edx
  cmp edx, eax
end;

//----- Haste on party with dead weak members

function HasteWeakCheck(p: int):LongBool;
begin
  Result:= (pint8(p + 8)^ <> 0) and (pint8(p + 8*14)^ = 0) and (pint8(p + 8*16)^ = 0);
end;

procedure HasteWeakHook1;
asm
  push eax
  push edx
  sub eax, 8
  call HasteWeakCheck
  test eax, eax
  pop edx
  pop eax
end;

procedure HasteWeakHook2;
asm
  lea eax, [edi - 8]
  call HasteWeakCheck
  test eax, eax
end;

//----- Report errors
// Delphi exception handling with DebugHook doesn't work here for some reason,
//  so I have to emulate exceptions.

procedure AppendDebugLog(const s:string);
const
  Sep = '================================================================================'#13#10#13#10;
begin
  RSAppendTextFile(AppPath + 'ErrorLog.txt', s + Sep);
end;

type
  TDelphiExceptionArgs = packed record
    RetAddr: ptr;
    ExceptObj: ptr;
    Ebx: int;
    Esi: int;
    Edi: int;
    Ebp: int;
    Esp: int;
  end;
  
  TNonDelphiExceptionArgs = packed record
    Context: PContext;
    ExceptObj: ptr;
  end;

  TNotifyTerminateArgs = packed record
    RetAddr: ptr;
  end;

const
  cDelphiException    = $0EEDFADE;
  cNonDelphiException = $0EEDFAE4;
  cDelphiTerminate    = $0EEDFAE2;

var
  MyRaiseExceptionProc: procedure(Code, Flags, ArgCount: int; Args: ptr); stdcall;
  LastUnhandledExceptionFilter: function(var p: TExceptionPointers): int; stdcall;

const
  EXCEPTION_EXECUTE_HANDLER = 1;
  EXCEPTION_CONTINUE_SEARCH = 0;
  EXCEPTION_CONTINUE_EXECUTION = -1;


function UnhandledException(var p: TExceptionPointers): int; stdcall;
type
  TExceptObjProc = function(p: Windows.PExceptionRecord): Exception;
var
  args: TNonDelphiExceptionArgs;
  args2: TNotifyTerminateArgs absolute args;
begin
  args.Context:= p.ContextRecord;
  args.ExceptObj:= TExceptObjProc(ExceptObjProc)(p.ExceptionRecord);
  MyRaiseExceptionProc(cNonDelphiException, 0, 2, @args);
  AppendDebugLog(RSLogExceptions);
  args2.RetAddr:= nil;
  MyRaiseExceptionProc(cDelphiTerminate, 0, 1, @args2);
  Exception(args.ExceptObj).Free;
  if @LastUnhandledExceptionFilter <> nil then
    Result:= LastUnhandledExceptionFilter(p)
  else
    Result:= EXCEPTION_CONTINUE_SEARCH;
end;


procedure ErrorProc(RetAddr: int; Reason: PChar; Ebx, Esi, Edi, Ebp, Esp, Ecx, Edx, Eax: int); stdcall;
var
  c: TContext;
  args: TNonDelphiExceptionArgs;
  args2: TNotifyTerminateArgs absolute args;
begin
  c.Ebx:= Ebx;
  c.Esi:= Esi;
  c.Edi:= Edi;
  c.Ebp:= Ebp;
  c.Esp:= Esp;
  c.Eax:= Eax;
  c.Ecx:= Ecx;
  c.Edx:= Edx;
  c.Eip:= RetAddr - 5;
  args.Context:= @c;
  if Reason = nil then
    args.ExceptObj:= Exception.Create('(internal error)')
  else
    args.ExceptObj:= Exception.Create(Reason);

  MyRaiseExceptionProc(cNonDelphiException, 0, 2, @args);
  AppendDebugLog(RSLogExceptions);
  args2.RetAddr:= ptr(RetAddr);
  MyRaiseExceptionProc(cDelphiTerminate, 0, 1, @args2);
  Exception(args.ExceptObj).Free;
end;

procedure ErrorHook1;
asm
  push eax
  lea eax, [esp + $C]
  push edx
  push ecx
  lea ecx, [eax + 4]
  push ecx             // esp
  push ebp
  push edi
  push esi
  push ebx
  push 0               // Reason
  push dword ptr [eax] // RetAddr
  call ErrorProc
  
  jmp [$4D8280]
end;

procedure ErrorHook2;
const
  TextBuffer = int(_TextBuffer1);
asm
  push eax
  lea eax, [esp + $18]
  push edx
  push ecx
  lea ecx, [eax + 4]
  push ecx             // esp
  push ebp
  push edi
  push esi
  push ebx
  push TextBuffer      // Reason
  push dword ptr [eax] // RetAddr
  call ErrorProc

  mov ecx, [$720808]
end;

procedure ErrorHook3;
asm
  push eax
  mov eax, ebp
  push edx
  mov edx, [esp + $10]
  push ecx
  lea ecx, [eax + 8]
  push ecx                   // esp
  push dword ptr [eax]       // ebp
  push dword ptr [eax - $2C] // edi
  push dword ptr [eax - $28] // esi
  push dword ptr [eax - $24] // ebx
  push edx                   // Reason
  push dword ptr [eax + 4]   // RetAddr
  call ErrorProc

  jmp [$4D826C]
end;

//----- MusicLoopsCount (main part is in KeysProc)

procedure ChangeTrackHook;
asm
  mov eax, MusicLoopsCount
  dec eax
  mov MusicLoopsLeft, eax
  mov LastMusicStatus, 0

  // std
  test byte ptr [$6BE1E4], $10
end;

//----- Fix chests: place items that were left over

procedure FixChest(p, chest: int);
var
  ItemsToPlace: array[0..139] of int;
  i, j, h: int;
begin
  h:= pint(_ChestWidth + 4*pword(p)^)^*pint(_ChestHeight + 4*pword(p)^)^ - 1;
  inc(p, 4);
  for i := 0 to 139 do
    ItemsToPlace[i]:= pint(p + _ItemOff_Size*i)^;
  inc(p, _ItemOff_Size*140);
  for i := 0 to h do
    if pint2(p + i*2)^ > 0 then
      ItemsToPlace[pint2(p + i*2)^ - 1]:= 0;
  for i := 0 to 139 do
    if ItemsToPlace[i] <> 0 then
      for j := 0 to h do
        if (pint2(p + j*2)^ = 0) and _Chest_CanPlaceItem(0, ItemsToPlace[i], j, chest) then
        begin
          _Chest_PlaceItem(0, i, j, chest);
          break;
        end;
end;

procedure FixChestHook;
asm
  test byte ptr [edi], 2
  jz @exit
  lea eax, [edi - 2]
  mov edx, ecx
  call FixChest
  push $42041C
@exit:
end;

//----- Limit blaster & bow speed with BlasterRecovery

procedure FixBlasterSpeed;
asm
  mov eax, ecx
  cmp eax, BlasterRecovery
  jnl @exit
  mov eax, BlasterRecovery
@exit:
end;

//----- Ignore DDraw errors

procedure DDrawErrorHook;
asm
  mov eax, esi
  pop esi
  pop ebp
  ret $10
end;

//----- Attacking big monsters D3D

var
  SpriteD3DPoints: array[0..9] of single;

function VisibleSpriteD3DProc(p: psingle; var count: int):psingle;
const
  vx1 = pint($F8BA94);
  vy1 = pint($F8BA98);
  vx2 = pint($F8BA9C);
  vy2 = pint($F8BAA0);
var
  i: int;
  x1, x2, y1, y2: single;
begin
  Result:= p;
  y1:= p^;
  y2:= p^;
  dec(p);
  x1:= p^;
  x2:= p^;
  inc(p, 8);
  for i:= 1 to 3 do
  begin
    if p^ < x1 then  x1:= p^;
    if p^ > x2 then  x2:= p^;
    inc(p);
    if p^ < y1 then  y1:= p^;
    if p^ > y2 then  y2:= p^;
    inc(p, 7);
  end;
  x1:= max(x1, vx1^ + 0.1);
  x2:= min(x2, vx2^ - 0.1);
  y1:= max(y1, vy1^ + 0.1);
  y2:= min(y2, vy2^ - 0.1);
  if (x1 > x2) or (y1 > y2) then
  begin
    count:= 3;  // just check 1 point
    exit;
  end;
  dec(count);  // 5 points instead of 4
  SpriteD3DPoints[0]:= (x1 + x2)/2;
  SpriteD3DPoints[1]:= (y1 + y2)/2;
  SpriteD3DPoints[2]:= x1;
  SpriteD3DPoints[3]:= y1;
  SpriteD3DPoints[4]:= x1;
  SpriteD3DPoints[5]:= y2;
  SpriteD3DPoints[6]:= x2;
  SpriteD3DPoints[7]:= y1;
  SpriteD3DPoints[8]:= x2;
  SpriteD3DPoints[9]:= y2;
  Result:= @SpriteD3DPoints[1];
end;

procedure VisibleSpriteD3DHook;
asm
  lea eax, dword ptr $EF5144[esi]
  lea edx, [ebp + $C]  // number of points
  jmp VisibleSpriteD3DProc
end;

//----- Credits move too fast

procedure FixCreditsPause;
asm
  push 12
  call dword ptr [$4D80B8] // Sleep
end;

//----- Fix party generation screen clouds & arrows speed

var
  DrawPartyScreenStd: procedure;
  LastPartyScreenDraw: uint;

procedure DrawPartyScreenHook;
var
  k: uint;
  o1, o2: int;
begin
  k:= timeGetTime;
  o1:= pint($A74B4C)^; o2:= pint($A74B54)^;
  DrawPartyScreenStd;
  if (k - LastPartyScreenDraw) < 15 then
  begin
    pint($A74B4C)^:= o1; pint($A74B54)^:= o2;
  end else
    LastPartyScreenDraw:= k;
end;

//----- No HWL for sprites

function Power2(n: int):int;
begin
  Result:= 1;
  while Result < n do
    Result:= Result*2;
end;

function FindSprite(Name: PChar): PSprite;
var
  i: int;
begin
  i:= pint(_SpritesLod + $EC9C)^;
  Result:= @Sprites;
  while (i > 0) and (_strcmpi(@Result.Name, Name) <> 0) do
  begin
    inc(Result);
    dec(i);
  end;
  if i = 0 then
    Result:= nil;
end;

procedure LoadPaletteD3D(var pal; PalIndex: int);
var
  i: int;
  p, p1: PWordArray;
begin
  PalIndex:= _LoadPalette(0, 0, $80D018, PalIndex);
  p:= @pal;
  p1:= ptr($8DE618 + PalIndex*32*256*2);
  for i := 0 to 255 do
    p[i]:= p1[i] and $1F + (p1[i] and $FFC0) shr 1 + $8000;
end;

function LoadSpriteD3DHook(Name: PChar; PalIndex: int): PHwlBitmap; stdcall;
var
  sprite: PSprite;
  i, j, x1, x2, y1, y2: int;
  pal: array[0..255] of word;
  p: pword;
begin
  Result:= nil;
  sprite:= FindSprite(Name);
  if sprite = nil then
    exit;

  with sprite^ do
  begin
    Result:= _new(SizeOf(THwlBitmap));
    //ZeroMemory(Result, SizeOf(THwlBitmap)); // now done always
    // Find area bounds
    x1:= w;
    x2:= -1;
    y1:= -1;
    y2:= -1;
    for i := 0 to h - 1 do
      with Lines[i] do
        if a1 >= 0 then
        begin
          if y1 < 0 then  y1:= i;
          y2:= i;
          if a1 < x1 then  x1:= a1;
          if a2 > x2 then  x2:= a2;
        end;
    with Result^ do
    begin
      FullW:= w;
      FullH:= h;
      if y1 < 0 then  exit;
      // Area dimensions must be powers of 2
      inc(x2);
      inc(y2);
      BufW:= Power2(x2 - x1);
      BufH:= Power2(y2 - y1);
      AreaW:= BufW;
      AreaH:= BufH;
      x1:= (x1 + x2 - BufW) div 2;
      //x1:= IntoRange(x1, 0, w - BufW);
      x2:= x1 + BufW - 1;
      y1:= (y1 + y2 - BufH) div 2;
      //y1:= IntoRange(y1, 0, h - BufH);
      y2:= y1 + BufH - 1;
      AreaX:= x1;
      AreaY:= y1;

      // Get Palette
      LoadPaletteD3D(pal, PalIndex);
      pal[0]:= 0;

      // Render
      Buffer:= _new(BufW*BufH*2);
      //ZeroMemory(Buffer, BufW*BufH*2); // now done always
      p:= Buffer;
      for i := y1 to y2 do
        with Lines[i] do
          if (i >= 0) and (i < h) and (a1 >= 0) then
          begin
            inc(p, a1 - x1);
            for j := 0 to a2 - a1 do
            begin
              p^:= pal[ord((pos + j)^)];
              inc(p);
            end;
            inc(p, x2 - a2);
          end else
            inc(p, BufW);
    end;
  end;
end;

procedure LoadSpriteD3DHook2;
asm
  inc dword ptr [esi+0EC9Ch]
  push $4ACA1E
  push $4A4FD8
  ret
end;

var
  HaveIt: Boolean; // found sprite by name, but palette don't match

procedure LoadSpriteD3DHook3;
asm
  mov eax, [esi+0EC9Ch]
  cmp HaveIt, 0
  jz @load
  mov HaveIt, 0
  pop ecx
  push $4AC996
@load:
end;

procedure LoadSpriteD3DHook4;
asm
  mov eax, [esi + 0ECB0h]
  mov ecx, [ebp - $10]  // sprite offset
  mov eax, dword ptr [eax + ecx + 4]  // sprite pal
  cmp eax, [ebp + $C]  // check pal
  jz @found
  cmp [esi + $EC9C], SpritesMax  // check for sprites overflow
  jz @found
  mov HaveIt, 1
  push $4AC76E
  ret
@found:
  mov HaveIt, 0
  push $4AC851
end;

//----- Extend Sprites Limits

procedure ExtendSpriteLimits;
const
  PatchPtr: array[0..25] of int = ($41E4BE, $41E9B9, $4403EB, $440DFB, $441BB9, $441BD1, $47AE29, $47B044, $47BC4B, $4A8FAC, $450CFB, $450D30, $46244A, $46250A, $462541, $462569, $4AC827, $4AC8DD, $4AC90F, $4AC94A, $4D7B50, $4AC8AC, $4AC8C1, $4AC8F7, $4AC96F, $4AC986);
  PatchCount: array[0..4] of int = ($462563, $462444, $4D7B46, $4AC7AD, $4AC782);
var
  Hook: TRSHookInfo;
  i: int;
begin
  ZeroMemory(@Hook, SizeOf(Hook));
  Hook.t:= RSht4;
  Hook.add:= int(@Sprites) - _SpritesOld;
  for i := 0 to high(PatchPtr) do
  begin
    Hook.p:= PatchPtr[i];
    RSApplyHook(Hook);
  end;
  Hook.add:= 0;
  for i := 0 to high(PatchCount) do
  begin
    Hook.p:= PatchCount[i];
    Hook.new:= (pint(Hook.p)^ div 1500) * SpritesMax;
    RSApplyHook(Hook);
  end;
end;

//----- Take sprite contour into account when clicking it

type
  TDrawSprite = packed record
    Texture: ptr;
    VertNum: int;
    Vert: array[0..3] of TD3DTLVertex;
    // ...
  end;

var
  SpriteD3DHitStd: function(var draw: TDrawSprite; x, y: single): LongBool; stdcall;

function FindSpriteD3D(texture: ptr): PSpriteD3D;
var
  i: int;
begin
  i:= pint(_SpritesLod + $EC9C)^ - 1;
  Result:= pptr(_SpritesLod + $ECB0)^;
  while (i > 0) and (Result.Texture <> texture) do
  begin
    inc(Result);
    dec(i);
  end;
  if i = 0 then
    Result:= nil;
end;

function SpriteD3DHitHook(var draw: TDrawSprite; x, y: single): LongBool; stdcall;
var
  sp3d: PSpriteD3D;
  sp: PSprite;
  drX, drW, drY, drH: Single;
  i, j: int;
begin
  Result:= SpriteD3DHitStd(draw, x, y);
  if not Result then
    exit;

  sp3d:= FindSpriteD3D(draw.Texture);
  if sp3d = nil then
    exit;
  with draw do
  begin
    drX:= Vert[0].sx;
    drW:= Vert[3].sx - drX;
    drY:= Vert[0].sy;
    drH:= Vert[1].sy - drY;
  end;
  i:= sp3d.AreaX + Round(sp3d.AreaW * (x - drX) / drW);
  j:= sp3d.AreaY + Round(sp3d.AreaH * (y - drY) / drH);
  sp:= FindSprite(sp3d.Name);
  if (sp = nil) or (sp.Lines = nil) then
    exit;

  Result:= false;
  if (j < 0) or (j >= sp.h) then  exit;
  with sp.Lines[j] do
  begin
    if (a1 < 0) or (i > a2) or (i < a1) then  exit;
    Result:= ((pos + i - a1)^ <> #0);
  end;
end;

//----- Use Delphi memory manager

function MemoryNewProc(n: uint): ptr;
begin
  Result:= AllocMem(n + 256); // +256 is to have some protection from out-of-bounds
  //ZeroMemory(PChar(Result) + n, 256);
  ZeroMemory(PChar(Result), n + 256);
end;

procedure MemoryFreeProc(p: ptr);
begin
  FreeMem(p);
end;
{
 // shows out-of-bounds problems
function MemoryNewProc(n: uint): ptr;
var
  i: uint;
begin
  i:= (n + $FFF) and not $FFF;
  Result:= VirtualAlloc(nil, i + $1000, MEM_RESERVE, PAGE_NOACCESS);
  VirtualAlloc(Result, i, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
  inc(PChar(Result), i - n);
end;

procedure MemoryFreeProc(p: ptr; ret: int);
begin
  //RSAppendTextFile('c:\palog.txt', IntToHex(int(p), 6) + #9 + IntToHex(ret, 6) + #13#10);
  //VirtualProtect(p, $1000, PAGE_NOACCESS, @p);
  VirtualFree(ptr(uint(p) and not $FFF), 0, MEM_RELEASE);
end;
{}

procedure MemoryInitHook;
asm
  mov [esi + $1D4D0], $10000000  // in Arena there's a check for available heap size
end;

procedure MemoryFreeHook;
asm
  mov eax, ebp
  call MemoryFreeProc
  push $4264EC
end;

procedure MemoryNewHook;
asm
  cmp eax, edi
  jge @ok
  xor eax, eax
  jmp @exit
@ok:
  call MemoryNewProc
@exit:
  push $426835
end;

//----- Fix general.txt parsing out-of-bounds

procedure GlobalTxtHook;
asm
  cmp [esp + $1C], $5E4A94
  jnl @exit
  push $4CC17B
@exit:
end;

//----- Fix spells.txt parsing out-of-bounds

procedure SpellsTxtHook;
asm
  cmp [esp + $1C], ($4E4416 - $14)
  jg @exit
  push $4CC17B
@exit:
end;

//----- Fix facet interception checking out-of-bounds

procedure FacetCheckHook;
asm
  cmp eax, 38
  jl @norm
  sub eax, 40
@norm:
  movzx esi, word ptr [eax+esi+2]
end;

procedure FacetCheckHook2;
asm
  cmp ecx, 38
  jl @norm
  sub ecx, 40
@norm:
  movzx esi, word ptr [ecx+esi+2]
end;

//----- There may be facets without vertexes

procedure NoVertexFacetHook;
asm
  mov al, [edi + $5d]
  test al, al
  jnz @norm
  pop eax
  push $49AE52
  ret
@norm:
  mov al, [edi + $5C]
  cmp al, 3
end;

//----- No HWL for bitmaps

function LoadBitmapD3DHook(n1, n2, this, PalIndex: int; Name: PChar): PHwlBitmap;
var
  bmp: TLodBitmap;
  f: ptr;
  pack: array of Byte;
  p: PWordArray;
  p1: PByteArray;
  pal: array[0..255] of Word;
  i, c: int;
begin
  Result:= nil;
  f:= _LodFind(0, 0, _BitmapsLod, 0, Name);
  if f = nil then
    exit;

  _fread(bmp, 1, $30, f); // read bitmap header
  Result:= _new(SizeOf(THwlBitmap));
  ZeroMemory(Result, SizeOf(THwlBitmap));
  with bmp, Result^ do
  begin
    FullW:= w;
    FullH:= h;
    AreaW:= w;
    AreaH:= h;
    BufW:= w;
    BufH:= h;
    Buffer:= _new(BmpSize*2);
    p:= Buffer;
    p1:= Buffer;
    // Read bitmap data
    if UnpSize <> 0 then
    begin
      SetLength(pack, DataSize);
      _fread(pack[0], 1, DataSize, f);
      _Deflate(0, @UnpSize, p1^, DataSize, pack[0]);
      pack:= nil;
    end else
      _fread(p1^, 1, DataSize, f);

    // Get Palette
    c:= 0;
    _fread(c, 1, 3, f);  // check first color
    LoadPaletteD3D(pal, Palette);
    if (c = $FFFF00) or (c = $FF00FF) or (c = $FC00FC) or (c = $FCFC00) then
      pal[0]:= 0;  // Margenta/light blue for transparency

    // Render
    for i := BmpSize - 1 downto 0 do
      p[i]:= pal[p1[i]];
  end;
end;

//----- Get rid of palettes limit in D3D
{
procedure PalettesHook;
asm
  cmp esi, $31
  jl @next
  push $48A5C5
  ret
@next:
  push $48A5B0
end;
}
//----- D3D better space reaction

const
  CVis_get_object_zbuf_val: function(n1, n2, this, obj: int): Word = ptr($4C1B1C);
var
  PressSpaceStd: function(n1, index, ref: int): Byte;

function PressSpaceHook: Byte;
var
  i, j, p: int;
begin
  p:= pint(pint($71FE94)^ + 3660)^;
  Result:= 1;
  for i := 0 to pint(p + 8200)^ - 1 do
  begin
    j:= CVis_get_object_zbuf_val(0, 0, p, pint(p + 6152 + i*4)^);
    Result:= PressSpaceStd(0, j shr 3, j);
    if Result = 0 then
      exit;
  end;
end;

//----- Correct door state switching: param = 3

procedure DoorStateSwitchHook;
asm
  shl ecx, 4
  add ecx, eax
  cmp ebx, 3
  jnz @exit
  mov ax, [ecx + $4C]
  xor ebx, ebx
  dec ax
  jz @exit
  dec ax
  jz @exit
  inc ebx
@exit:
end;

//----- Mouse look

const
  MLSideX = 640 - (8 + 468);
  MLSideY = 480 - (8 + 351);
var
  MiddleX, MiddleY, MLookPartX, MLookPartY: int;
  MWndPos, MLastPos, MLookTempPos: TPoint;
  MLookStartTime: DWORD;
  EmptyCur, ArrowCur: HCURSOR;
  MouseLookOn, MLookIsTemp: Boolean;

function CursorPos:PPoint;
begin
  Result:= PPoint(PPChar($720808)^ + $108);
end;

function GetMLookPoint(var p: TPoint): BOOL;
begin
  if MouseLookOn then
  begin
    CursorPos^:= Point(MiddleX, MiddleY);
    p:= MWndPos;
    ClientToScreen(_MainWindow^, p);
    Result:= true;
  end else
    Result:= GetCursorPos(p);
end;

procedure CheckMouseLook;
const
  myAnd: int = -1;
  myXor: int = 0;
var
  cur: HCURSOR;
  r: TRect;
begin
  GetClientRect(_MainWindow^, r);
  if _ScreenW^ <> 640 then
  begin
    // compatibility with resolution patches like mmtool's one
    MiddleX:= (r.Right - MLSideX) div 2;
    MiddleY:= (r.Bottom - MLSideY) div 2;
    MWndPos.X:= MiddleX;
    MWndPos.Y:= MiddleY;
  end else
  begin
    // compatibility with Angel's resolution mod
    MiddleX:= (640 - MLSideX) div 2;
    MiddleY:= (480 - MLSideY) div 2;
    MWndPos.X:= (MiddleX*r.Right + 639) div 640;
    MWndPos.Y:= (MiddleY*r.Bottom + 479) div 480;
  end;

  if ArrowCur = 0 then
    ArrowCur:= LoadCursor(GetModuleHandle(nil), 'Arrow');
  if EmptyCur = 0 then
    EmptyCur:= CreateCursor(GetModuleHandle(nil), 0, 0, 1, 1, @myAnd, @myXor);
  cur:= GetClassLong(_MainWindow^, -12);
  with Options do
    MouseLookOn:= MouseLook and (_CurrentScreen^ = 0) and (_MainMenuCode^ = -1) and
       ((cur = EmptyCur) or (cur = ArrowCur)) and
       ( (GetAsyncKeyState(MouseLookTempKey) and $8000 = 0) xor
          MouseLookChanged xor MouseLookUseAltMode xor
          (CapsLockToggleMouseLook and (GetKeyState(VK_CAPSLOCK) and 1 <> 0)));

  if MouseLookOn <> (cur = EmptyCur) then
  begin
    if not MouseLookOn then
    begin
      SetClassLong(_MainWindow^, -12, ArrowCur);
      _NeedRedraw^:= 1;
    end else
      SetClassLong(_MainWindow^, -12, EmptyCur);

    if MLookIsTemp or not MouseLookOn and (GetTickCount - MLookStartTime < 10*1000) then
    begin
      MLastPos:= MLookTempPos;
      MLookIsTemp:= false;
    end else
    begin
      GetMLookPoint(MLastPos);
      if MouseLookOn then
      begin
        MLookIsTemp:= GetAsyncKeyState(Options.MouseLookTempKey) and $8000 <> 0;
        GetCursorPos(MLookTempPos);
        MLookStartTime:= GetTickCount;
      end;
    end;

    SetCursorPos(MLastPos.X, MLastPos.Y);
  end;
end;

procedure ProcessMouseLook;

  function Partial(move: int; var part: int; factor: int): int;
  var
    x: int;
  begin
    x:= part + move*factor;
    Result:= (x + 32) and not 63;
    part:= x - Result;
    Result:= Result div 64;
  end;

const
  dir = _Party_Direction;
  angle = _Party_Angle;
  AngleLim = 180;
var
  p: TPoint;
  speed: PPoint;
begin
  CheckMouseLook;
  GetCursorPos(p);
  if MouseLookOn and (GetForegroundWindow = _MainWindow^) then
  begin
    if MLookIsTemp then
      speed:= @MLookSpeed2
    else
      speed:= @MLookSpeed;
    dir^:= (dir^ - Partial((p.X - MLastPos.X), MLookPartX, speed.X)) and 2047;
    angle^:= IntoRange(angle^ - Partial((p.Y - MLastPos.Y), MLookPartY, speed.Y),
      -Options.MaxMLookAngle, Options.MaxMLookAngle);
    if (p.X <> MLastPos.X) or (p.Y <> MLastPos.Y) then
    begin
      p:= MWndPos;
      ClientToScreen(_MainWindow^, p);
      SetCursorPos(p.X, p.Y);
    end;
  end;
  MLastPos:= p;
end;

procedure MouseLookHook(p: TPoint); stdcall;
begin
  CheckMouseLook;
  if MouseLookOn then
    CursorPos^:= Point(MiddleX, MiddleY)
  else
    CursorPos^:= p;
end;

function MouseLookHook2(var p: TPoint): BOOL; stdcall;
begin
  CheckMouseLook;
  Result:= GetMLookPoint(p);
end;

var
  MouseLookHook3Std: procedure(a1, a2, this: ptr);
  MLookBmp: TBitmap;

procedure MLookLoad;
const
  CurFile = 'Data\MouseLookCursor.bmp';
var
  b: TBitmap;
  exist: Boolean;
begin
  MLookBmp:= TBitmap.Create;
  with MLookBmp, Canvas do
  begin
    PixelFormat:= pf16bit;
    HandleType:= bmDIB;
    b:= nil;
    exist:= FileExists(CurFile);
    if exist then
      try
        b:= TBitmap.Create;
        b.LoadFromFile(CurFile);
        Width:= b.Width;
        Height:= b.Height;
        CopyRect(ClipRect, b.Canvas, ClipRect);
        b.Free;
        exit;
      except
        RSShowException;
        b.Free;
      end;

    Width:= 64;
    Height:= 64;
    Brush.Color:= $F0A0B0;
    FillRect(ClipRect);
    DrawIconEx(Handle, 32, 32, ArrowCur, 0, 0, 0, 0, DI_DEFAULTSIZE or DI_NORMAL);
    if not exist then
      try
        SaveToFile(CurFile);
      except
      end;
  end;
end;

procedure MLookDraw;
var
  p1, p2: PChar;
  x, y, w, h, d1, d2: int;
  k, trans: Word;
begin
  if MLookBmp = nil then
    MLookLoad;
  w:= MLookBmp.Width;
  h:= MLookBmp.Height;
  p1:= MLookBmp.ScanLine[0];
  d1:= PChar(MLookBmp.ScanLine[1]) - p1 - 2*w;
  p2:= PPChar($E31B54)^ + 2*(_ScreenW^*(MiddleY - h div 2) + MiddleX - w div 2);
  d2:= 2*(_ScreenW^ - w);
  trans:= pword(p1)^;
  for y := 1 to h do
  begin
    for x := 1 to w do
    begin
      k:= pword(p1)^;
      if k <> trans then
        pword(p2)^:= k;
      inc(p1, 2);
      inc(p2, 2);
    end;
    inc(p1, d1);
    inc(p2, d2);
  end;
end;

procedure MouseLookHook3(a1, a2, this: ptr);
begin
  CheckMouseLook;
  if MouseLookOn and not MLookIsTemp then
    MLookDraw;

  MouseLookHook3Std(nil, nil, this);
end;

//----- A beam of Prismatic Light in the center of screen that doesn't disappear

var
  FixPrismaticBugStd: function(size, unk: uint):ptr cdecl;

function FixPrismaticBug(size, unk: uint):ptr; cdecl;
begin
  Result:= FixPrismaticBugStd(size, unk);
  ZeroMemory(Result, size);
end;

//----- Window procedure hook

procedure WindowProcStdImpl;
asm
  push ebp
  mov ebp, esp
  sub esp, $48
  push $46382E
end;

var
  WindowProcStd: function(w: HWND; msg: uint; wp: WPARAM; lp: LPARAM):HRESULT; stdcall;

function WindowProcHook(w: HWND; msg: uint; wp: WPARAM; lp: LPARAM):HRESULT; stdcall;
const
  CommandsArray = $721458;
  AddCommand: procedure(a1, a2, this, cmd: int) = ptr($4760C5);
begin
  Result:= WindowProcStd(w, msg, wp, lp);
  if (msg = WM_MOUSEWHEEL) and Options.MouseWheelFly then
    if wp < 0 then
      AddCommand(0, 0, CommandsArray, 14)
    else
      AddCommand(0, 0, CommandsArray, 13);
end;

//----- Fly in Z axis with mouse

procedure MouseLookFlyProc(var v: TPoint; ebp: PChar);
const
  MaxH = 4000;
  PartyH = pint($ACCE3C);
  PartyZ = pint($ACD4F4);
  DriftZ = pint($ACD528);
  PartyState = pint($AD45B0);
  z = -$20;
  z2 = -$48;
  HasCeil = -$74;
  CeilH = -$58;
var
  x, y, a, b: ext;
  dz: int;
begin
  if (v.X = 0) and (v.Y = 0) then  exit;
  with Options do
    if (_Flying^ = 0) or not MouseLook or not MouseFly then
      exit;

  a:= _Party_Direction^*pi/1024;
  x:= cos(a);
  y:= sin(a);
  a:= x*v.X + y*v.Y;
  b:= _Party_Angle^*pi/1024;
  dz:= Round(a*sin(b)*_TimeDelta^/128);
  if (dz > 0) and (pint(ebp + z)^ + dz > MaxH) then
    dz:= 0;
  a:= a*(cos(b) - 1);
  v.X:= v.X + Round(x*a);
  v.Y:= v.Y + Round(y*a);
  inc(pint(ebp + z)^, dz);
  inc(pint(ebp + z2)^, dz);
  if (pint(ebp + HasCeil)^ <> 0) and (pint(ebp + z)^ < pint(ebp + CeilH)^) and
     (pint(ebp + z)^ + PartyH^ >= pint(ebp + CeilH)^) then
  begin
    PartyZ^:= pint(ebp + CeilH)^ - PartyH^ - 1;
    pint(ebp + z)^:= PartyZ^;
    pint(ebp + z2)^:= PartyZ^;
    DriftZ^:= PartyZ^;
    pint($721458)^:= 0;
    _Flying^:= 0;
    PartyState^:= PartyState^ or 1;
    v.X:= 0;
    v.Y:= 0;
  end;
end;

procedure MouseLookFlyHook1;
asm
  mov ds:[$ACD4F8], eax
  push edi
  push ebx
  mov eax, esp
  mov edx, ebp
  call MouseLookFlyProc
  pop ebx
  pop edi
  mov ds:[ebp - $C], edi
end;

procedure MouseLookFlyHook2;
asm
  push ebx
  push edi
  mov eax, esp
  mov edx, ebp
  call MouseLookFlyProc
  pop edi
  pop ebx
  mov eax, [ebp - 4]
  xor ecx, ecx
end;

//----- Show resistances of monster

var
  IDMonStr: string;

function IDMonProc(n: int): PChar;
begin
  IDMonStr:= IntToStr(n);
  Result:= PChar(IDMonStr);
end;

procedure IDMonHook;
asm
  call IDMonProc
  xchg eax, [esp]
  jmp eax
end;

//----- negative/0 causes a crash in stats screen

procedure StatColorFixHook;
asm
  test esi, esi
  jnz @norm
  dec esi
@norm:
  idiv esi
  mov ecx, $FF
end;

//----- Pause the game in Town Portal screen

var
  TPFixStd: procedure;

procedure TPFixHook;
begin
  _PauseTime;
  TPFixStd;
end;

//----- Use Smooth turn rate by default

procedure DefaultSmoothTurnHook;
asm
  mov edx, 3
  mov ecx, $4E46C8
end;

//----- Waiting used to recover characters twise as fast

procedure FixWaitHook;
asm
  test dword ptr ds:[$ACD6BC], 2
  jz @skip
  sar dword ptr ds:[esp + 4], 1

@skip:
  push $48E8ED
end;

//----- Was no LeaveMap event on death

procedure FixLeaveMapDieHook;
const
  OnMapLeave: int = $443FB8;
asm
  call OnMapLeave
  push $4CAC80
end;

//----- Was no LeaveMap event with walk travel

procedure FixLeaveMapWalkHook;
const
  OnMapLeave: int = $443FB8;
asm
  call OnMapLeave
  mov ecx, $576CB0
end;

//----- Subtract gold after autosave when trevelling

var
  TravelGold: int;

procedure TravelGoldFixHook1;
asm
  mov TravelGold, ecx
  xor ecx, ecx
  push $492BAE
end;

procedure TravelGoldFixHook2;
asm
  mov eax, TravelGold
  sub ds:[$ACD56C], eax
  push $4B1DF5
end;

//----- Switch to 16 bit color when going windowed

var
  AutoColor16Std, AutoColor16Std2: ptr;

procedure AutoColor16Proc;
var
  mode: TDevMode;
begin
  if (GetDeviceCaps(GetDC(0), BITSPIXEL) <> 16) or (GetDeviceCaps(GetDC(0), PLANES) <> 1) then
    with mode do
    begin
      dmSize:= SizeOf(mode);
      dmBitsPerPel:= 16;
      dmFields:= DM_BITSPERPEL;
      RSWin32Check(ChangeDisplaySettings(mode, CDS_FULLSCREEN) = DISP_CHANGE_SUCCESSFUL);
    end;
end;

procedure AutoColor16Hook;
asm
  push AutoColor16Std
  jmp AutoColor16Proc
end;

procedure AutoColor16Hook2;
asm
  push ecx
  call AutoColor16Proc
  mov ecx, AutoColor16Std2
  xchg ecx, [esp]
end;

//----- Show accurate errors of D3DInit

procedure BetterD3DInitErrorsHook;
asm
  mov eax, [esi + 40088h]
  lea eax, [eax + 48h]
  xchg eax, [esp]
  jmp eax
end;

//----- Autorun key like in WoW

procedure AutorunProc;
type
  PQueue = ^TQueue;
  TQueue = packed record
    n: int;
    key: array[0..29] of int;
  end;
const
  Queue = PQueue($721458);
  Add: procedure(n1,n2, this, key: int) = ptr($4760C5);
var
  i: int;
begin
  Autorun:= Autorun and (_CurrentScreen^ = 0) and not _TurnBased^;
  if not Autorun or (_Paused^ <> 0) then  exit;
  for i:= 0 to Queue.n - 1 do
    if Queue.key[i] in [3..6, 16, 17] then
    begin
      Autorun:= false;
      exit;
    end;
  Add(0, 0, int(Queue), 16);
end;

procedure AutorunHook;
asm
  push $4304D6
  jmp AutorunProc
end;

//----- Lloyd: take spell points and action after autosave

procedure LloydAutosaveFix;
asm
  // get back spell points and remove recovery delay
  push esi
  mov esi, [esp + $600 - $5E8]
  mov ax, [esi + _CharOff_Recover]
  push eax
  mov word ptr [esi + _CharOff_Recover], 0
  mov eax, [esi + _CharOff_SpellPoints]
  push eax
  add eax, dword ptr [$5061C4]
  mov [esi + _CharOff_SpellPoints], eax

  call _DoSaveGame

  // restore decreased spell points and recovery delay
  pop eax
  mov [esi + _CharOff_SpellPoints], eax
  pop eax
  mov [esi + _CharOff_Recover], ax
  pop esi
end;

//----- TP: take action after autosave

procedure TPAutosaveFix;
asm
  // remove recovery delay
  push edx
  xor eax, eax
  mov al, [$50BF40]
  mov edi, _CharOff_Size
  mul eax, edi
  pop edx
  lea edi, [eax + $ACD804 + _CharOff_Recover]
  mov ax, [edi]
  push eax
  mov word ptr [edi], 0

  call _DoSaveGame

  // restore recovery delay
  pop eax
  mov [edi], ax
  mov edi, 1
end;

//----- Fix crash when looking too low (lower than game allows)

procedure FixSkyCrash;
asm
  sar edx, $10
  cmp [ebp-$C], 0
  jg @ok  // unexpected case
  mov ecx, edx
  sal ecx, 1
  cmp ecx, [ebp-$C]
  jg @ok
  mov eax, $40000000 // -1
  ret
@ok:
  idiv [ebp-$C]
end;

procedure FixDivCrash1;
asm
  // std code
  sar edx, 15  // 15 instead of 16, cause it's idiv, EDX must be 2 times smaller 
  // (edx < C) <> (-edx < C) means overflow, equality means overflow
  cmp edx, [ebp - $10]
  jg @g
  jz @bad
  // edx < C
  neg edx
  cmp edx, [ebp - $10]
  jl @good
  jmp @bad
@g: // edx > C
  neg edx
  cmp edx, [ebp - $10]
  jle @bad

@good:
  neg edx
  sar edx, 1
  // std code
  idiv [ebp - 10h]
  ret

@bad:
  // return $7FFFFFFF or -$7FFFFFFF
  mov eax, [ebp - $C]
  xor eax, [ebp - $10]
  shr eax, 31
  add eax, $7FFFFFFF
  or eax, 1
end;

//----- Custom LODs - find file

type
  TCustomLod = record
    Std: ptr;
    Mine: ptr;
  end;
  TLodRecord = record
    Lod: ptr;
    PtrPassedAsName: ptr;
    Name: array[0..$3F] of char;
  end;
  TLodRecords = record
    Index: int;
    Records: array[0..255] of TLodRecord;
  end;

var
  CustomLods: array of TCustomLod;
  LodRecords: TLodRecords;

const
  FindInLodPtr2 = $461659 + 5;

procedure AddLodRecord(ALod: ptr; AName: PChar);
begin
  LodRecords.Index:= uint(LodRecords.Index - 1) and 255;
  with LodRecords, Records[Index] do
  begin
    Lod:= ALod;
    PtrPassedAsName:= AName;
    StrLCopy(@Name, AName, $39);
  end;
end;

procedure FindInLodAndSeek;
asm
  push ebp
  mov ebp, esp
  push ebx
  push esi
  jmp eax
end;

procedure FindInLod;
asm
  // remove Unsorted param from stack
  pop eax
  pop edx
  mov [esp], edx
  push eax
  // std code
  push ebx
  push esi
  push edi
  mov esi, ecx
  push FindInLodPtr2
end;

function FindInLodProc(StdPtr, _2: int; This: ptr; Unsorted: LongBool; Name: PChar): int;
var
  find: function(StdPtr, _2: int; This: ptr; Unsorted: LongBool; Name: PChar): int;
  i: int;
begin
  if StdPtr = FindInLodPtr2 then
    @find:= @FindInLod
  else
    @find:= @FindInLodAndSeek;

  for i := high(CustomLods) downto 0 do
    if CustomLods[i].Std = This then
    begin
      Result:= find(StdPtr, 0, CustomLods[i].Mine, Unsorted, Name);
      if Result <> 0 then
      begin
        AddLodRecord(CustomLods[i].Mine, Name);
        exit;
      end;
    end;
  Result:= find(StdPtr, 0, This, Unsorted, Name);
  if Result <> 0 then
    AddLodRecord(This, Name);
end;

procedure FindInLodAndSeekHook;
asm
  pop eax
  jmp FindInLodProc
end;

procedure FindInLodHook;
asm
  pop eax
  mov edx, [esp]
  push edx  // return address
  mov edx, [esp+8]
  mov [esp+4], edx  // Name
  jmp FindInLodProc
end;

//----- Custom LODs - load

const
  LodSize = $23C;
  LodChapOff = $214;
  LodLoad: function(_1, _2: int; This: ptr;
    CanWrite: int; Name: PChar): LongBool = ptr($461812);
  LodLoadChapter: function(_1, _2: int; This: ptr;
    Chapter: PChar): LongBool = ptr($4618C7);
  LodClean: procedure(_1, _2: int; This: ptr) = ptr($46146E);

function DoLoadCustomLod(Old: PChar; Name, Chap: PChar): ptr;
var
  i: int;
begin
  GetMem(Result, LodSize);
  if SameText(Chap, 'chapter') then
    Chap:= 'maps';
  if LodLoad(0,0, Result, 0, Name) or LodLoadChapter(0,0, Result, Chap) then
  begin
    LodClean(0,0, Result);
    FreeMem(Result, LodSize);
    Result:= nil;
  end else
  begin
    if SameText(Chap, 'maps') then
      StrCopy(Old + LodChapOff, 'chapter');

    i:= length(CustomLods);
    SetLength(CustomLods, i + 1);
    CustomLods[i].Std:= Old;
    CustomLods[i].Mine:= Result;
  end;
end;

function LoadCustomLod(Old: PChar; Name: PChar): ptr; stdcall;
begin
  Result:= DoLoadCustomLod(Old, Name, Old + LodChapOff);
end;

function FreeCustomLod(Lod: ptr): Boolean; stdcall;
var
  i: int;
begin
  for i := high(CustomLods) downto 0 do
    if CustomLods[i].Mine = Lod then
    begin
      ArrayDelete(CustomLods, i, SizeOf(CustomLods[0]));
      SetLength(CustomLods, length(CustomLods) - 1);
      LodClean(0,0, Lod);
      FreeMem(Lod, LodSize);
      Result:= true;
      exit;
    end;
  Result:= false;
end;

function GetCustomLodsList: ptr; stdcall;
begin
  Result:= ptr(CustomLods);
end;

function GetLodRecords: ptr; stdcall;
begin
  Result:= @LodRecords;
end;

procedure LoadCustomLods(Old: int; Name: string; Chap: PChar);
begin
  with TRSFindFile.Create('Data\*.' + Name) do
    try
      while FindNextAttributes(0, FILE_ATTRIBUTE_DIRECTORY) do // Only files
        DoLoadCustomLod(ptr(Old), PChar('Data\' + Data.cFileName), Chap);
    finally
      Free;
    end;
end;

var
  LoadLodsOld: TProcedure;

// OnStart, just before loading any data
procedure LoadLodsHook;
begin
  LoadCustomLods($6D0490, 'icons.lod', 'icons');
  LoadCustomLods($6BE8D8, 'events.lod', 'icons');
  LoadCustomLods($6F0D00, 'bitmaps.lod', 'bitmaps');
  LoadCustomLods($6E2048, 'sprites.lod', 'sprites08');
  LoadCustomLods($6A08E0, 'games.lod', 'chapter');
  LoadLodsOld;
end;

//----- Called whenever time is resumed

procedure ClearKeyStatesHook;
var
  i: int;
begin
  for i := 1 to 255 do
    MyGetAsyncKeyState(i);
  GetCursorPos(MLastPos);
end;

//----- Save game bug in Windows Vista and higher (bug of OS or other software)

procedure SaveGameBugHook(old, new: PChar); cdecl;
begin
  while not MoveFile(old, new) do
  begin
    Sleep(1);
    DeleteFileA(new);
  end;
end;

//----- Strafe in MouseLook

function StrafeOrWalkHook(key: int): bool; stdcall;
begin
  Result:= MyGetAsyncKeyState(key) <> 0;
  if AlwaysStrafe or MouseLookOn then
    Result:= not Result;
end;

//----- Fix movement rounding problems

type
  TStrafeRemainder = array[0..4] of int;
var
  CurRemainder: pint;
  PlayerStrafe: TStrafeRemainder;
  MonsterStrafe: array[0..499] of TStrafeRemainder;
  ObjectStrafe: array[0..999] of TStrafeRemainder;
  OldMoveStructInit: ptr;

procedure FixStrafe1;
asm
  imul [ebp - 4]
  mov ecx, CurRemainder
  add eax, [ecx]
  adc edx, 0
  mov [ecx], ax
  add ecx, 4
  mov CurRemainder, ecx
  shrd eax, edx, $10
end;

procedure FixStrafe2;
asm
  imul dword ptr [$72142C]
  add eax, $8000
  adc edx, 0
end;

procedure FixStrafePlayer;
asm
  lea eax, PlayerStrafe
  mov CurRemainder, eax
  jmp OldMoveStructInit
end;

procedure FixStrafeMonster1;
asm
  mov eax, [ebp - 4]
  lea eax, [eax*4 + eax]
  lea eax, MonsterStrafe[eax*4]
  mov CurRemainder, eax
  jmp OldMoveStructInit
end;

procedure FixStrafeMonster2;
asm
  mov eax, [ebp - 8]
  lea eax, [eax*4 + eax]
  lea eax, MonsterStrafe[eax*4]
  mov CurRemainder, eax
  jmp OldMoveStructInit
end;

procedure FixStrafeObject1;
asm
  mov eax, [ebp - $1C]
  lea eax, [eax*4 + eax]
  lea eax, ObjectStrafe[eax*4]
  mov CurRemainder, eax
  jmp OldMoveStructInit
end;

procedure FixStrafeObject2;
asm
  mov eax, [ebp - $20]
  lea eax, [eax*4 + eax]
  lea eax, ObjectStrafe[eax*4]
  mov CurRemainder, eax
  jmp OldMoveStructInit
end;

//----- Control D3D water

function HDWTRCountHook(time: uint): int;
begin
  Result:= (time div HDWTRDelay) mod HDWTRCount;
end;

//----- Use antialiasing and anisatropic filtration (doesn't work)

procedure BetterD3dProc;
var
  dev: ^IDirect3DDevice3;
begin
  dev:= ptr(pint($E31AF0)^ + 56);
  dev^.SetRenderState(D3DRENDERSTATE_ANTIALIAS, ord(D3DANTIALIAS_SORTINDEPENDENT));
  dev^.SetRenderState(D3DRENDERSTATE_EDGEANTIALIAS, 1);
  dev^.SetRenderState(D3DRENDERSTATE_TEXTUREMAG, uint(D3DFILTER_LINEARMIPLINEAR));
end;

procedure BetterD3dHook;
asm
  call dword ptr [ecx+0A0h]
  push $4A09A6
  jmp BetterD3dProc
end;

//----- Buka localization

var
  _sprintfex: ptr;

//----- HooksList

var
  HooksList: array[1..179] of TRSHookInfo = (
    (p: $45B0D1; newp: @KeysHook; t: RShtCall; size: 6), // My keys handler
    (p: $4655FE; old: $452C75; backup: @@SaveNamesStd; newp: @SaveNamesHook; t: RShtCall), // Buggy autosave file name localization
    (p: $45E5A4; old: $45E2D0; backup: @FillSaveSlotsStd; newp: @FillSaveSlotsHook; t: RShtCall), // Fix Save/Load Slots
    (p: $45EB36; old: $45E2D0; newp: @FillSaveSlotsHook; t: RShtCall), // Fix Save/Load Slots
    (p: $4352CD; old: $4309C7; backup: @ChooseSaveSlotBackup; newp: @ChooseSaveSlotHook; t: RSht4), // Fix Save/Load Slots
    (p: $43520D; old: $430B16; backup: @SaveBackup; newp: @SaveHook; t: RSht4), // Fix Save/Load Slots
    (p: $460002; newp: @QuicksaveHook; t: RShtCall), // Quicksave trigger
    (p: $45E33F; newp: @QuickSaveSlotHook; t: RShtCall), // Show multiple Quick Saves
    (p: $45E65B; newp: @QuickSaveNamesHook; t: RShtCall; size: 6), // Quick Saves names
    (p: $4606FD; newp: @QuickSaveDrawHook; t: RShtCall; size: 6), // Quick Saves names Draw
    (p: $4605F2; newp: @QuickSaveDrawHook2; t: RShtCall; size: 6), // Quick Saves names Draw
    (p: $45F84E; old: $461B85; backup: @@SaveSetPauseStd; newp: @SaveSetPauseHook; t: RShtCall), // Autosave didn't pause time
    (p: $42FC46; newp: @CapsLockHook; t: RShtCall; size: 6; Querry: 4), // CapsLockToggleRun
    (p: $4A95C6; old: $4A9550; new: $4A9595; t: RSht4; Querry: 3), // No death movie
    (p: $4A95B2; old: $4A950F; new: $4A9595; t: RSht4; Querry: 2), // No intro: 3dologo
    (p: $4A95B6; old: $4A951C; new: $4A9595; t: RSht4; Querry: 2), // No intro: new world logo
    (p: $4A95BA; old: $4A9529; new: $4A9595; t: RSht4; Querry: 2), // No intro: jvc
    (p: $4A95C2; old: $4A9536; new: $4A9595; t: RSht4; Querry: 2), // No intro: Intro
    (p: $4A953C; newp: @IntroHook; t: RShtCall; size: 7; Querry: 2), // No intro: Intro Post
    (p: $4A95BE; old: $4A9543; new: $4A9536; t: RSht4; Querry: 2), // No intro: Intro Post - include Intro too
    (p: $4651F0; new: $465241; t: RShtJmp; size: 6; Querry: 1), // NoCD
    (p: $4AC2D7; old: $840F; new: $E990; t: RSht2), // Fix XP compatibility
    (p: $474CEF; old: $F5; new: 0; t: RSht4), // Fix Walk Sound
    (p: $47371E; old: $9A; new: 0; t: RSht4), // Fix Walk Sound
    (p: $4BF572; old: $72; new: $EB; t: RSht1), // Mok's patch smack bug
    (p: $44686D; newp: @InactiveMemberEventsHook; t: RShtCall; size: 6), // Multiple fauntain drinks bug
    (p: $493891; newp: @CheckNextMemberHook; t: RShtCall; size: 8; Querry: 8), // Switch between inactive members in inventory screen
    (p: $4938B5; newp: @CheckNextMemberHook; t: RShtCall; size: 8; Querry: 8), // Switch between inactive members in inventory screen
    (p: $418361; old: $418437; newp: @AttackDescriptionHook; t: RShtJmp), // Show recovery time for Attack
    (p: $418387; old: $418437; newp: @ShootDescriptionHook; t: RShtJmp), // Show recovery time for Shoot
    (p: $41697B; old: 500; new: 600; t: RSht4; Querry: 9), // Allow using Harden Item on artifacts
    (p: $4167FC; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416814; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416837; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416845; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41684E; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416857; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41686C; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416878; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $4168F7; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416905; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41690E; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416917; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41692C; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416938; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416961; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41696A; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $416974; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $41697F; old: $416773; new: $41677E; t: RShtJmp6), // Don't waste potion bottle if it has no effect
    (p: $4CAD70; newp: @_sprintfex; newref: true; t: RShtJmp; size: 6; Querry: 6), // Buka localization
    (p: $4BE7C1; old: 1000; new: 1; t: RSht4; Querry: 5), // Delay before showing video
    (p: $4BEB54; old: 300; new: 1; t: RSht4; Querry: 5), // Delay after showing video
    (p: $410897; newp: @LodFilesHook; t: RShtCall; size: 6; Querry: 12), // Load files from DataFiles folder
    (p: $48D021; old: $45827D; newp: @SecondHandDaggerHook; t: RShtCall), // Dagger tripple damage from 2nd hand check
    (p: $48E431; old: $45827D; newp: @WeaponRecoveryBonusHook; t: RShtCall; size: 8), // Include items bonus to skill in recovery time
    (p: $48E43B; t: RShtNop; size: 2), // Include items bonus to skill in recovery time
    (p: $4685ED; old: $492C03; backup: @ScrollTestMemberStd; newp: @ScrollTestMember; t: RShtCall), // Don't allow using scrolls by not active member
    (p: $41AB2E; old: $495446; backup: @ReputationHookStd; newp: @ReputationHook; t: RShtCall; Querry: 7), // Show numerical reputation value
    (p: $41AB3C; old: $4E3010; newp: PChar('%s: '#12'%05d%d %s'#12'00000'); t: RSht4; Querry: 7), // Show numerical reputation value
    (p: $41AB48; old: $14; new: $18; t: RSht1; Querry: 7), // Show numerical reputation value
    (p: $492EA0; newp: ptr($492DD3); t: RShtJmp; size: $492EBA - $492EA0), // Poison condition ignored Protection from Magic
    (p: $426376; old: $4262C0; backup: @DoubleSpeedStd; newp: @DoubleSpeedHook; t: RShtCall), // DoubleSpeed
    (p: $472C6F; old: $4CA74C; backup: @TurnSpeedStd; newp: @TurnSpeedHook; t: RShtCall), // TurnSpeed
    (p: $472C97; old: $4CA74C; newp: @TurnSpeedHook; t: RShtCall), // TurnSpeed
    (p: $473DA9; old: $4CA74C; newp: @TurnSpeedHook; t: RShtCall), // TurnSpeed
    (p: $473DD1; old: $4CA74C; newp: @TurnSpeedHook; t: RShtCall), // TurnSpeed
    (p: $474F92; old: $48E4F0; backup: @FeatherFallStd; newp: @FeatherFallHook; t: RShtCall), // fix 'of Feather Falling' items
    (p: $48CF05; newp: @DaggerTrippleHook; t: RShtCall; Querry: 10), // ProgressiveDaggerTrippleDamage
    (p: $48D033; newp: @DaggerTrippleHook; t: RShtCall; Querry: 10), // ProgressiveDaggerTrippleDamage
    (p: $4B6B10; old: 1500; newp: @HorsemanSpeakTime; newref: true; t: RSht4), // Horseman delay
    (p: $4B6B23; old: 2500; newp: @BoatmanSpeakTime; newref: true; t: RSht4), // Boatman delay
    (p: $42934D; newp: @HasteWeakHook1; t: RShtCall), // Haste on party with dead weak members
    (p: $42D9AD; newp: @HasteWeakHook2; t: RShtCall), // Haste on party with dead weak members
    (p: $466B9B; newp: @ErrorHook1; t: RShtCall; size: 6), // Report errors
    (p: $466C0C; newp: @ErrorHook2; t: RShtCall; size: 6), // Report errors
    (p: $466E5D; newp: @ErrorHook3; t: RShtCall; size: 6), // Report errors
    (p: $4AA0CF; newp: @ChangeTrackHook; t: RShtCall; size: 7), // MusicLoopsCount
    (p: $432AC8; old: $7B; new: $71; t: RSht1), // Fix menu return
    (p: $420412; newp: @FixChestHook; t: RShtCall; Querry: 11), // Fix chests: place items that were left over
    (p: $48E4E3; newp: @FixBlasterSpeed; t: RShtCall; size: 6), // Limit blaster & bow speed with BlasterRecovery
    (p: $466CD8; newp: @DDrawErrorHook; t: RShtJmp), // Ignore DDraw errors
    (p: $42262C; old: $D75; new: $22EB; t: RSht2), // Remove code left from MM6 (pretty harmless, but still a bug)
    (p: $4C0A23; newp: @VisibleSpriteD3DHook; t: RShtCall; size: 6), // Attacking big monsters D3D
    (p: $4C0B31; old: $20; new: 8; t: RSht1), // Attacking big monsters D3D
    (p: $4A4FF1; old: $452504; newp: @LoadSpriteD3DHook; t: RShtCall), // No HWL for sprites
    (p: $4ACA13; old: $4A4FD8; newp: @LoadSpriteD3DHook2; t: RShtJmp), // No HWL for sprites
    (p: $4AC8CD; old: $4AC98A; newp: @LoadSpriteD3DHook3; t: RShtCall), // No HWL for sprites
    (p: $4AC768; old: $4AC851; newp: @LoadSpriteD3DHook4; t: RShtJmp6), // No HWL for sprites
    (p: $49EAE8; size: 16), // No HWL for sprites
    (p: $49EB49; size: 5), // No HWL for sprites
    (p: $4C1508; old: $4C1579; backup: @@SpriteD3DHitStd; newp: @SpriteD3DHitHook; t: RShtCall), // Take sprite contour into account when clicking it
    (p: $4C115A; old: $4C1235; new: $4C1386; t: RShtCall), // Never can be sure a sprite obscures a facet
    (p: $4C115F; size: 8), // Never can be sure a sprite obscures a facet
    (p: $4C151E; old: $4C088F; new: $4C0C96; t: RShtCall), // Don't check sprite visibility when clicking
    (p: $4C1523; size: 4), // Don't check sprite visibility when clicking
    (p: $497E2B; newp: @FixCreditsPause; t: RShtCall; size: 6), // Credits move too fast
    (p: $4975E9; old: $495B4F; backup: @@DrawPartyScreenStd; newp: @DrawPartyScreenHook; t: RShtCall), // Fix party generation screen clouds & arrows speed
    (p: $426689; newp: @MemoryInitHook; t: RShtCall; size: 8), // Use Delphi memory manager
    (p: $426446; newp: @MemoryFreeHook; t: RShtJmp; size: 6), // Use Delphi memory manager
    (p: $42674E; newp: @MemoryNewHook; t: RShtJmp; size: 7), // Use Delphi memory manager
    (p: $452D3A; old: $4CC17B; newp: @GlobalTxtHook; t: RShtCall), // Fix global.txt parsing out-of-bounds
    (p: $453B11; old: $4CC17B; newp: @SpellsTxtHook; t: RShtCall), // Fix spells.txt parsing out-of-bounds
    (p: $453EEC; old: 29; new: 28; t: RSht4), // Fix history.txt parsing out-of-bounds
    (p: $475A67; newp: @FacetCheckHook; t: RShtCall), // Fix facet ray interception checking out-of-bounds
    (p: $475B42; newp: @FacetCheckHook; t: RShtCall), // Fix facet ray interception checking out-of-bounds
    (p: $475C15; newp: @FacetCheckHook; t: RShtCall), // Fix facet ray interception checking out-of-bounds
    (p: $4C21AB; newp: @FacetCheckHook2; t: RShtCall), // Fix facet interception checking out-of-bounds
    (p: $4C22A9; newp: @FacetCheckHook2; t: RShtCall), // Fix facet interception checking out-of-bounds
    (p: $4C239D; newp: @FacetCheckHook2; t: RShtCall), // Fix facet interception checking out-of-bounds
    (p: $49AC92; newp: @NoVertexFacetHook; t: RShtCall), // There may be facets without vertexes
    (p: $4A4D93; old: $452504; newp: @LoadBitmapD3DHook; t: RShtCall; Querry: 13), // No HWL for bitmaps
    (p: $49EAE3; old: $4523AB; new: $45246B; t: RShtCall; Querry: 13), // No HWL for bitmaps
    (p: $49EB3B; size: 5; Querry: 13), // No HWL for bitmaps
    //(p: $48A5B9; newp: @PalettesHook; t: RShtJmp), // Get rid of palettes limit in D3D
    (p: $46A364; newp: ptr($46A38C); t: RShtJmp), // Ignore 'Invalid ID reached!'
    (p: $44EC22; newp: ptr($44EC4A); t: RShtJmp), // Ignore 'Sprite outline currently Unsupported'
    (p: $46A1A9; old: $46A338; backup: @@PressSpaceStd; newp: @PressSpaceHook; t: RShtCall), // D3D better space reaction
    (p: $449AC3; newp: @DoorStateSwitchHook; t: RShtCall), // Correct door state switching: param = 3
    (p: $46A08E; newp: @MouseLookHook; t: RShtJmp; size: 10), // Mouse look
    (p: $4D825C; newp: @MouseLookHook2; t: RSht4), // Mouse look
    (p: $4160C5; backup: @@MouseLookHook3Std; newp: @MouseLookHook3; t: RShtCall), // Mouse look
    (p: $42FE25; newp: @StrafeOrWalkHook; t: RShtCall; size: 6; Querry: 15), // Strafe in MouseLook
    (p: $42FEA5; newp: @StrafeOrWalkHook; t: RShtCall; size: 6; Querry: 15), // Strafe in MouseLook
    (p: $4CB071; backup: @@FixPrismaticBugStd; newp: @FixPrismaticBug; t: RShtCall), // A beam of Prismatic Light in the center of screen that doesn't disappear
    (p: $463828; newp: @WindowProcHook; t: RShtJmp; size: 6), // Window procedure hook
    (p: $474300; newp: @MouseLookFlyHook1; t: RShtCall), // Fix strafes and walking rounding problems
    (p: $47303C; newp: @MouseLookFlyHook2; t: RShtCall), // Fix strafes and walking rounding problems
    (p: $41F312; newp: @IDMonHook; t: RShtCall; size: 6), // Show resistances of monster
    (p: $4178C5; newp: @StatColorFixHook; t: RShtCall; size: 7), // negative/0 causes a crash in stats screen
    (p: $45A994; newp: @FixKeyConfigHook; t: RShtCall), // Fix keys configuration loading
    (p: $4D8868; newp: @PaletteSMul; newref: true; t: RSht4), // Control palette gamma
    (p: $4D886C; newp: @PaletteVMul; newref: true; t: RSht4), // Control palette gamma
    (p: $411C1E; old: $411AB6; backup: @@TPFixStd; newp: @TPFixHook; t: RShtCall), // Pause the game in Town Portal screen
    (p: $465B9E; newp: @DefaultSmoothTurnHook; t: RShtCall), // Use Smooth turn rate by default
    (p: $493917; old: $48E8ED; newp: @FixWaitHook; t: RShtCall), // Waiting used to recover characters twise as fast
    (p: $4636E4; old: $4CAC80; newp: @FixLeaveMapDieHook; t: RShtCall), // Was no LeaveMap event on death
    (p: $432E63; newp: @FixLeaveMapWalkHook; t: RShtCall), // Was no LeaveMap event with walk travel
    (p: $4D8260; newp: @MyGetAsyncKeyState; t: RSht4), // Don't rely on bit 1 of GetAsyncKeyState
    (p: $4B6997; old: $492BAE; newp: @TravelGoldFixHook1; t: RShtCall), // Subtract gold after autosave when trevelling
    (p: $4B6AF4; old: $4B1DF5; newp: @TravelGoldFixHook2; t: RShtCall), // Subtract gold after autosave when trevelling
    (p: $46532F; t: RShtNop; size: $11), // Switch to 16 bit color when going windowed
    (p: $49DE88; old: $4CB9C0; backup: @AutoColor16Std; newp: @AutoColor16Hook; t: RShtCall), // Switch to 16 bit color when going windowed
    (p: $4A0987; old: $4A11AC; backup: @AutoColor16Std2; newp: @AutoColor16Hook2; t: RShtCall), // Switch to 16 bit color when going windowed
    (p: $4A0662; newp: @BetterD3DInitErrorsHook; t: RShtCall), // Show accurate errors of D3DInit
    (p: $46334B; old: $4304D6; newp: @AutorunHook; t: RShtCall), // Autorun key like in WoW
    (p: $43367E; newp: @LloydAutosaveFix; t: RShtCall), // Lloyd: take spell points and action after autosave
    (p: $43393B; newp: @TPAutosaveFix; t: RShtCall), // TP: take action after autosave
    (p: $4338CB; t: RShtNop; size: 5), // Town Portal triggered autosave even within a location
    (p: $44F0A8; new: $44F100; t: RShtJmp), // My bug: quicksave set to F11 not working
    (p: $48FDF2; old: 3; new: 2; t: RSht1), // Spear skill wasn't added to damage on expert level (only on master)
    (p: $466CC6; newp: ptr($10C2); t: RSht4), // Ignore DD errors
    (p: $49A5D9; old: 0; new: 1; t: RSht1), // Fix DLV search in games.lod
    (p: $47E60A; old: 0; new: 1; t: RSht1), // Fix DDM search in games.lod
    (p: $4798BA; newp: @FixSkyCrash; t: RShtCall; size: 6), // Fix crash when looking too low (lower than game allows)
    (p: $485220; newp: @FixDivCrash1; t: RShtCall; size: 6), // Fix int overflow crash in editor
    (p: $4615BD; newp: @FindInLodAndSeekHook; t: RShtCall), // Custom LODs
    (p: $461659; newp: @FindInLodHook; t: RShtCall), // Custom LODs
    (p: $4655FE; backup: @@LoadLodsOld; newp: @LoadLodsHook; t: RShtCall), // Custom LODs
    (p: $459E78; newp: @ClearKeyStatesHook; t: RShtJmp; size: 6), // Clear my keys as well
    (p: $461EFF; newp: @SaveGameBugHook; t: RShtCall), // Save game bug in Windows Vista and higher (bug of OS or other software)
    (p: $470578; newp: @FixStrafe1; t: RShtCall; size: 7), // Fix movement rounding problems
    (p: $4705A4; newp: @FixStrafe1; t: RShtCall; size: 7), // Fix movement rounding problems
    (p: $4705CB; newp: @FixStrafe1; t: RShtCall; size: 7), // Fix movement rounding problems
    (p: $4705F2; newp: @FixStrafe1; t: RShtCall; size: 7), // Fix movement rounding problems
    (p: $470614; newp: @FixStrafe1; t: RShtCall; size: 7), // Fix movement rounding problems
    (p: $46FDAF; backup: @OldMoveStructInit; newp: @FixStrafeMonster1; t: RShtCall), // Fix movement rounding problems
    (p: $470C44; newp: @FixStrafeMonster2; t: RShtCall), // Fix movement rounding problems
    (p: $471690; newp: @FixStrafeObject1; t: RShtCall), // Fix movement rounding problems
    (p: $471E5A; newp: @FixStrafeObject2; t: RShtCall), // Fix movement rounding problems
    (p: $473161; newp: @FixStrafePlayer; t: RShtCall), // Fix movement rounding problems
    (p: $4745B9; newp: @FixStrafePlayer; t: RShtCall), // Fix movement rounding problems
    (p: $4731D1; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $4731F1; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $473214; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $474633; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $474656; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $474679; newp: @FixStrafe2; t: RShtCall; size: 6), // Fix movement rounding problems
    (p: $441252; newp: @HDWTRCountHook; t: RShtCall; size: $441282 - $441252), // Control D3D water
    (p: $465EBB; old: 7; newp: @HDWTRCount; newref: true; t: RSht1), // Control D3D water

//    (p: $4A0951; newp: @BetterD3dHook; t: RShtJmp; size: 6),
    ()
  );

procedure ReadDisables;
var
  i: int;
begin
  with TIniFile.Create(AppPath+'mm7.ini') do
    try
      for i := low(HooksList) to high(HooksList) do
        if ReadBool('Disable', 'Hook'+IntToStr(i), false) then
          HooksList[i].Querry:= -100;
    finally
      Free;
    end
end;

procedure HookAll;
var
  LastDebugHook: DWord;
begin
  Assert(RSCheckHooks(HooksList), SWrong);

  ExtendSpriteLimits;

  ReadDisables;
  RSApplyHooks(HooksList);
  if NoCD and FileExists('Anims\Magic7.vid') then
    RSApplyHooks(HooksList, 1);
  if NoIntro then
    RSApplyHooks(HooksList, 2);
  if NoDeathMovie then
    RSApplyHooks(HooksList, 3);
  if not CapsLockToggleRun then
    RSApplyHooks(HooksList, 4);
  if NoVideoDelays then
    RSApplyHooks(HooksList, 5);
  if UseMM7text and (RSLoadProc(_sprintfex, AppPath + 'mm7text.dll', '_sprintfex') <> 0) then
    RSApplyHooks(HooksList, 6);
  if ReputationNumber then
    RSApplyHooks(HooksList, 7);
  if FreeTabInInventory then
    RSApplyHooks(HooksList, 8);
  if HardenArtifacts then
    RSApplyHooks(HooksList, 9);
  if ProgressiveDaggerTrippleDamage then
    RSApplyHooks(HooksList, 10);
  if FixChests then
    RSApplyHooks(HooksList, 11);
  if DataFiles then
    RSApplyHooks(HooksList, 12);
  if NoBitmapsHwl then
    RSApplyHooks(HooksList, 13);
  if not StandardStrafe then
    RSApplyHooks(HooksList, 15);
  if PlayMP3 then
    HookMP3;

  RSDebugUseDefaults;
  LastDebugHook:= DebugHook;
  DebugHook:= 0;
  RSDebugHook(true);
  @MyRaiseExceptionProc:= RaiseExceptionProc;
  RSDebugHook(false);
  DebugHook:= LastDebugHook;
  @LastUnhandledExceptionFilter:= SetUnhandledExceptionFilter(@UnhandledException);
end;

exports
  LoadCustomLod,
  FreeCustomLod,
  GetCustomLodsList,
  GetLodRecords;
initialization
  @PlaySound:= @DoPlaySound;
  @WindowProcStd:= @WindowProcStdImpl;
finalization
  if EmptyCur <> 0 then
    DestroyCursor(EmptyCur);
end.
